#!/usr/bin/env python
# coding=utf-8
# ==============================================================================
# title           : PyNM.py
# description     : Gaussian Processes, Centiles & LOESS-based normative models
# author          : Guillaume Dumas (Institut Pasteur/Université de Montréal)
#                   Annabelle Harvey (Université de Montréal)
# date            : 2021-04-15
# notes           : The input dataframe column passed to --group must either
#                   have controls marked as "CTR" and probands as "PROB", or
#                   controls marked as 0 and probands as 1.
#                   The --pheno_p is for the path to the input dataframe.
#                   The --out_p flag is for the path to save the output
#                   dataframe, including filename formatted as 'filename.csv'.
#                   The confounds columns for the gaussian process model must
#                   be specified using the --confounds flag. The confound for
#                   the LOESS and centiles models must be specified using the
#                   --conf flag.
# licence         : BSD 3-Clause License
# python_version  : 3.7
# ==============================================================================

import pandas as pd
import numpy as np
import warnings
import matplotlib.pyplot as plt
import seaborn as sns
import warnings

from sklearn import gaussian_process
from sklearn.gaussian_process.kernels import Matern, WhiteKernel, ConstantKernel
from sklearn.model_selection import KFold
from statsmodels.stats.diagnostic import het_white
from statsmodels.tools.tools import add_constant
from scipy import stats

from pynm.util import *
from pynm.models.loess import *
from pynm.models.centiles import *

class PyNM:
    """ Class to run normative modeling using LOESS, centiles, or GP model.

    Attributes
    ----------
    data : dataframe
        Dataset to fit model, must at least contain columns corresponding to 'group',
        'score', and 'conf'.
    score : str
        Label of column from data with score (response variable).
    group : str
        Label of column from data that encodes wether subjects are probands or controls.
    CTR : str or int
        Label of controls in 'group' column can be 'CTR' or 0.
    PROB: str or int
        Label of probands in 'group' column can be 'PRB' or 1.
    conf: str
        Label of column from data with confound to use for LOESS and centiles models.
    confounds: list of str
        List of labels of columns from data with confounds. For GP model all confounds will be used,
        for LOESS and Centiles models only the first is used. For GAMLSS all confounds are used
        unless formulas are specified. Categorical values must be denoted by c(var) ('c' must be lower case).
    train_sample: str or float
        Which method to use for a training sample, can be 'controls' to use all the controls, 
        'manual' to be manually set, or a float in (0,1] for a percentage of controls.
    bin_spacing: int
        Distance between bins for LOESS & centiles models.
    bin_width: float
        Width of bins for LOESS & centiles models.
    bins: array
        Bins for the centiles and LOESS models.
    bin_count: array
        Number of controls in each bin.
    zm: array
        Mean of each bin (LOESS).
    zstd: array
        Standard deviation of each bin (LOESS).
    zci: array
        Confidence interval of each bin (LOESS).
    z: array
        Centiles for each bin.
    RMSE_LOESS: float
        RMSE of LOESS normative model
    SMSE_LOESS: float
        SMSE of LOESS normative model
    RMSE_Centiles: float
        RMSE of Centiles normative model
    SMSE_Centiles: float
        SMSE of Centiles normative model
    RMSE_GP: float
        RMSE of Gaussian Process normative model
    SMSE_GP: float
        SMSE of Gaussian Process normative model
    MSLL_GP: float
        MSLL of Gaussian Process normative model
    RMSE_GAMLSS: float
        RMSE of GAMLSS
    SMSE_GAMLSS: float
        SMSE of GAMLSS
    MSLL_GAMLSS: float
        MSLL of GAMLSS
    """

    def __init__(self, data, score, group, confounds, 
                train_sample=1, bin_spacing=-1, bin_width=-1, seed=None):
        """ Create a PyNM object.

        Parameters
        ----------
        data : dataframe
            Dataset to fit model, must at least contain columns corresponding to 'group',
            'score', and 'conf'.
        score : str
            Label of column from data with score (response variable).
        group : str
            Label of column from data that encodes wether subjects are probands or controls.
        confounds: list of str
            List of labels of columns from data with confounds. For GP model all confounds will be used,
            for LOESS and Centiles models only the first is used. For GAMLSS all confounds are used
            unless formulas are specified. Categorical values must be denoted by c(var) ('c' must be lower case).
        train_sample: str or float, default=1
            Which method to use for a training sample, can be a float in (0,1] for a percentage of controls 
            or 'manual' to be manually set using a column of the DataFrame labelled 'train_sample'.
        bin_spacing: int, default=-1
            Distance between bins for LOESS & centiles models.
        bin_width: float, default=-1
            Width of bins for LOESS & centiles models.
        seed: int, default=None
            Seed for random state generator, if None no seed is set.
        
        Raises
        ------
        ValueError
            Each row of DataFrame must have a unique index.
        """
        if data.index.nunique() != data.shape[0]:
            raise ValueError('Each row of DataFrame must have a unique index.')
        self.data = data.copy()
        self.score = score
        self.group = group
        self.confounds = confounds
        self.conf = self.confounds[0]
        self.train_sample = train_sample
        self.CTR = None
        self.PROB = None
        self.bin_spacing = bin_spacing
        self.bin_width = bin_width
        self.bins = None
        self.bin_count = None
        self.zm = None
        self.zstd = None
        self.zci = None
        self.z = None
        self.RMSE_LOESS = None
        self.SMSE_LOESS = None
        self.RMSE_Centiles = None
        self.SMSE_Centiles = None
        self.RMSE_GP = None
        self.SMSE_GP = None
        self.MSLL_GP = None
        self.RMSE_GAMLSS = None
        self.SMSE_GAMLSS = None
        self.MSLL_GAMLSS = None

        if seed is not None:
            np.random.seed(seed)

        self._set_group_names()
        self._set_group()

    def _make_train_sample(self, train_size):
        """ Select a subsample of controls to be used as a training sample for the normative model.

        Parameters
        ----------
        train_size: float
            Percentage of controls to use for training. Must be in (0,1].
        """
        ctr_idx = self.data[self.data[self.group] == self.CTR].index.tolist()
        n_ctr = len(ctr_idx)
        n_ctr_train = max(int(train_size*n_ctr), 1)

        np.random.seed(1)
        ctr_idx_train = np.array(np.random.choice(ctr_idx, size=n_ctr_train, replace=False))
        
        train_sample = np.zeros(self.data.shape[0])
        train_sample[ctr_idx_train] = 1
        self.data['train_sample'] = train_sample

        print('Models will be fit with train sample size = {}: using {}/{} of controls.'.format(train_size, n_ctr_train, n_ctr))

    def _set_group(self):
        """ Read the specified training sample and set the group attribute to refer to the appropriate column of data.

        Raises
        ------
        ValueError
            With train_sample=1: Dataset has no controls for training sample.
        ValueError
            With train_sample='manual': Data has no column "train_sample". To manually specify a training sample, 
            data .csv must contain a column "train_sample" with included subjects marked with 1 and rest as 0.
        ValueError
            With train_sample='manual': Dataset has no subjects in specified training sample.
        ValueError
            Value for train_sample not recognized. Must be either a value in (0,1] or 'manual'.
        ValueError
            With train_sample float: Numerical value for train_sample must be in the range (0,1].
        """
        if self.train_sample == 1:
            print('Models will be fit on full set of controls.')
            if self.data[self.data[self.group] == self.CTR].shape[0] == 0:
                raise ValueError('Dataset has no controls for training sample.')
        elif self.train_sample == 'manual':
            print('Models will be fit using specified training sample.')
            if 'train_sample' not in self.data.columns:
                raise ValueError('Data has no column "train_sample". To manually specify a training sample, data .csv '
                                 'must contain a column "train_sample" with included subjects marked with 1 and rest as 0.')
            self.group = 'train_sample'
            self._set_group_names()

            if self.data[self.data[self.group] == self.CTR].shape[0] == 0:
                raise ValueError('Dataset has no subjects in specified training sample..')
        else:
            try:
                train_size = float(self.train_sample)
            except:
                raise ValueError("Value for train_sample not recognized. Must be either 'controls', 'manual', or a "
                                 "value in (0,1].")
            else:
                if (train_size > 1) or (train_size <= 0):
                    raise ValueError("Numerical value for train_sample must be in the range (0,1].")
                else:
                    self._make_train_sample(train_size)
                    self.group = 'train_sample'
                    self._set_group_names()

    def _set_group_names(self):
        """ Read whether subjects in data are labeled CTR/PROB or 0/1 and set labels accordingly."""
        if self.group == 'train_sample':
            self.CTR = 1
            self.PROB = 0
        else:
            labels = list(self.data[self.group].unique())
            if ('CTR' in labels) or ('PROB' in labels):
                self.CTR = 'CTR'
                self.PROB = 'PROB'
            else:
                self.CTR = 0
                self.PROB = 1

    def _get_masks(self):
        """ Get masks from data corresponding to controls and probands.

        Returns
        -------
        array
            Control mask: controls marked as True.
        array
            Proband mask: probands marked as True.
        """
        ctr = self.data.loc[(self.data[self.group] == self.CTR)]
        ctr_mask = self.data.index.isin(ctr.index)
        probands = self.data.loc[(self.data[self.group] == self.PROB)]
        prob_mask = self.data.index.isin(probands.index)
        return ctr_mask, prob_mask

    # Default values for age in days
    def _create_bins(self):
        """ Create bins for the centiles and LOESS models.
        Returns
        -------
        array
            Bins for the centiles and LOESS models.
        """
        min_conf = self.data[self.conf].min()
        max_conf = self.data[self.conf].max()

        if self.bin_width == -1:
            self.bin_width = (max_conf - min_conf)/100
        if self.bin_spacing == -1:
            self.bin_spacing = (max_conf - min_conf)/10

        # define the bins (according to width)
        self.bins = np.arange(min_conf, max_conf + self.bin_width, self.bin_spacing)
        return self.bins

    def bins_num(self):
        """ Give the number of ctr used for the age bin each participant is in.

        Returns
        -------
        array
            Number of controls in each bin.
        """
        if self.bins is None:
            self.create_bins()

        dists = [np.abs(conf - self.bins) for conf in self.data[self.conf]]
        idx = [np.argmin(d) for d in dists]
        n_ctr = [self.bin_count[i] for i in idx]
        self.data['participants'] = n_ctr
        return n_ctr

    def _loess_rank(self):
        """ Associate ranks to LOESS normative scores."""
        self.data.loc[(self.data.LOESS_z <= -2), 'LOESS_rank'] = -2
        self.data.loc[(self.data.LOESS_z > -2) &
                      (self.data.LOESS_z <= -1), 'LOESS_rank'] = -1
        self.data.loc[(self.data.LOESS_z > -1) &
                      (self.data.LOESS_z <= +1), 'LOESS_rank'] = 0
        self.data.loc[(self.data.LOESS_z > +1) &
                      (self.data.LOESS_z <= +2), 'LOESS_rank'] = 1
        self.data.loc[(self.data.LOESS_z > +2), 'LOESS_rank'] = 2

    def loess_normative_model(self,cv_folds=1):
        """ Compute LOESS normative model.
        
        Parameters
        ----------
        cv_folds: int, default=1
            How many folds of cross-validation to perform. If 1, there is no cross-validation.
        """
        if self.bins is None:
            self._create_bins()
        
        # Format data
        data = self.data[[self.conf, self.score]].to_numpy(dtype=np.float64)

        # Take the controls
        ctr_mask, _ = self._get_masks()
        ctr = data[ctr_mask]

        # Cross-validation
        if cv_folds == 1:
            self.zm,self.zstd,self.zci = loess_fit(ctr,self.bins,self.bin_width)
            m, std = loess_predict(data,self.bins,self.zm,self.zstd)

            rmse = RMSE(self.data[self.score].values[ctr_mask],m[ctr_mask])
            smse = SMSE(self.data[self.score].values[ctr_mask],m[ctr_mask])
        
        else:
            kf = KFold(n_splits=cv_folds, shuffle=True)
            rmse = []
            smse = []
            print(f'Starting {cv_folds} folds of CV...')
            for i, (train_index, test_index) in enumerate(kf.split(ctr)):
                ctr_train, ctr_test = ctr[train_index], ctr[test_index]
                cv_zm,cv_zstd,_ = loess_fit(ctr_train,self.bins,self.bin_width)
                cv_m, _ = loess_predict(ctr_test,self.bins,cv_zm,cv_zstd)
                r = RMSE(ctr_test[:,1],cv_m)
                s = SMSE(ctr_test[:,1],cv_m)
                print(f'CV Fold {i}: RMSE={r:.3f} - SMSE={s:.3f}')
                rmse.append(r)
                smse.append(s)
            print('Done!')

            rmse = np.mean(rmse)
            smse = np.mean(smse)
            print(f'Average: RMSE={rmse:.3f} - SMSE={smse:.3f}')

            self.zm,self.zstd,self.zci = loess_fit(ctr,self.bins,self.bin_width)
            m, std = loess_predict(data,self.bins,self.zm,self.zstd)

        self.data['LOESS_pred'] = m
        self.data['LOESS_sigma'] = std
        self.data['LOESS_residuals'] = self.data[self.score] - self.data['LOESS_pred']
        self.data['LOESS_z'] = self.data['LOESS_residuals']/self.data['LOESS_sigma']

        self.RMSE_LOESS = rmse
        self.SMSE_LOESS = smse

        self._loess_rank()

    def _centiles_rank(self):
        """ Associate ranks to centiles associated with normative modeling."""
        self.data.loc[(self.data.Centiles_pred <= 5), 'Centiles_rank'] = -2
        self.data.loc[(self.data.Centiles_pred > 5) &
                      (self.data.Centiles_pred <= 25), 'Centiles_rank'] = -1
        self.data.loc[(self.data.Centiles_pred > 25) &
                      (self.data.Centiles_pred <= 75), 'Centiles_rank'] = 0
        self.data.loc[(self.data.Centiles_pred > 75) &
                      (self.data.Centiles_pred <= 95), 'Centiles_rank'] = 1
        self.data.loc[(self.data.Centiles_pred > 95), 'Centiles_rank'] = 2

    def centiles_normative_model(self, cv_folds=1):
        """ Compute centiles normative model.

        Parameters
        ----------
        cv_folds: int, default=1
            How many folds of cross-validation to perform. If 1, there is no cross-validation.
        """
        if self.bins is None:
            self._create_bins()

        # Format data
        data = self.data[[self.conf, self.score]].to_numpy(dtype=np.float64)

        # Take the controls
        ctr_mask, _ = self._get_masks()
        ctr = data[ctr_mask]

        # Cross-validation
        if cv_folds == 1:
            self.z = centiles_fit(ctr,self.bins,self.bin_width)
            result, centiles = centiles_predict(data,self.bins,self.z)
            centiles_50 = np.array([centiles[i, 50] for i in range(self.data.shape[0])])

            rmse = RMSE(self.data[self.score].values[ctr_mask],centiles_50[ctr_mask])
            smse = SMSE(self.data[self.score].values[ctr_mask],centiles_50[ctr_mask])
        
        else:
            kf = KFold(n_splits=cv_folds, shuffle=True)
            rmse = []
            smse = []
            print(f'Starting {cv_folds} folds of CV...')
            for i, (train_index, test_index) in enumerate(kf.split(ctr)):
                ctr_train, ctr_test = ctr[train_index], ctr[test_index]
                cv_z = centiles_fit(ctr_train,self.bins,self.bin_width)
                _, cv_centiles = centiles_predict(ctr_test, self.bins,cv_z)
                cv_50 = np.array([cv_centiles[i, 50] for i in range(ctr_test.shape[0])])
                r = RMSE(ctr_test[:,1],cv_50)
                s = SMSE(ctr_test[:,1],cv_50)
                print(f'CV Fold {i}: RMSE={r:.3f} - SMSE={s:.3f}')
                rmse.append(r)
                smse.append(s)
            print('Done!')

            rmse = np.mean(rmse)
            smse = np.mean(smse)
            print(f'Average: RMSE={rmse:.3f} - SMSE={smse:.3f}')

            self.z = centiles_fit(ctr,self.bins,self.bin_width)
            result, centiles = centiles_predict(data,self.bins,self.z)

        self.data['Centiles'] = result
        self.data['Centiles_5'] = np.array([centiles[i, 5] for i in range(self.data.shape[0])])
        self.data['Centiles_32'] = np.array([centiles[i, 32] for i in range(self.data.shape[0])])
        self.data['Centiles_pred'] = np.array([centiles[i, 50] for i in range(self.data.shape[0])])
        self.data['Centiles_68'] = np.array([centiles[i, 68] for i in range(self.data.shape[0])])
        self.data['Centiles_95'] = np.array([centiles[i, 95] for i in range(self.data.shape[0])])
        self.data['Centiles_sigma'] = (self.data['Centiles_68'] - self.data['Centiles_32'])/2
        self.data['Centiles_residuals'] = self.data[self.score] - self.data['Centiles_pred']
        self.data['Centiles_z'] = self.data['Centiles_residuals']/self.data['Centiles_sigma']

        self.RMSE_Centiles = rmse
        self.SMSE_Centiles = smse

        self._centiles_rank()

    def _get_conf_mat(self):
        """ Get confounds properly formatted from dataframe and input list.

        Returns
        -------
        array
            Confounds with categorical values dummy encoded. Dummy encoding keeps k-1
            dummies out of k categorical levels.
        """
        conf_clean, conf_cat = read_confounds(self.confounds)
        conf_mat = pd.get_dummies(self.data[conf_clean], columns=conf_cat, 
                                  drop_first=True)
        return conf_mat.to_numpy()

    def _get_score(self):
        """ Get the score from the PyNM object as an array.

        Raises
        ------
        ValueError
            Method must be one of "auto","approx", or "exact".

        Returns
        -------
        array
            The column of data marked by the user as 'score'.
        """
        return self.data[self.score].to_numpy()

    def _use_approx(self, method='auto'):
        """ Choose wether or not to use SVGP model. If method is set to 'auto' SVGP is chosen
        for datasets with more than 2000 points.

        Parameters
        ----------
        method: str, default='auto'
            Which method to use, can be 'exact' for exact GP regression, 'approx' for SVGP,
            or 'auto' which will set the method according to the size of the data.
        
        Raises
        ------
        ValueError
            Method must be one of "auto","approx", or "exact".
        """
        if method == 'auto':
            if self.data.shape[0] > 2000:
                return True
            else:
                return False
        elif method == 'approx':
            return True
        elif method == 'exact':
            if self.data.shape[0] > 2000:
                warnings.warn("Exact GP model with over 2000 data points requires "
                              "large amounts of time and memory, continuing with exact model.",Warning)
            return False
        else:
            raise ValueError('Method must be one of "auto","approx", or "exact".')
    
    def _test_gp_residuals(self,conf_mat):
        #Test normal
        k2, p_norm = stats.normaltest(self.data['GP_residuals'])
        if p_norm < 0.05:
            warnings.warn("The residuals are not Gaussian!")
        
        # Test heteroskedasticity
        exog = add_constant(conf_mat)
        _,p_het,_,_ = het_white((self.data['GP_residuals'])**2,exog)
        if p_het < 0.05:
            warnings.warn("The residuals are heteroskedastic!")
        
    def gp_normative_model(self, length_scale=1, nu=2.5, length_scale_bounds=(1e-5,1e5),method='auto', batch_size=256, n_inducing=500, num_epochs=20, cv_folds=1):
        """ Compute gaussian process normative model. Gaussian process regression is computed using
        the Matern Kernel with an added constant and white noise. For Matern kernel see scikit-learn documentation:
        https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.Matern.html.

        Parameters
        -------
        length_scale: float, default=1
            Length scale parameter of Matern kernel.
        nu: float, default=2.5
            Nu parameter of Matern kernel.
        length_scale_bounds: pair of floats >= 0 or 'fixed', default=(1e-5, 1e5)
            The lower and upper bound on length_scale. If set to 'fixed', ‘length_scale’ cannot be changed during hyperparameter tuning.
        method: str, default='auto'
            Which method to use, can be 'exact' for exact GP regression, 'approx' for SVGP,
            or 'auto' which will set the method according to the size of the data.
        batch_size: int, default=256
            Batch size for SVGP model training and prediction.
        n_inducing: int, default=500
            Number of inducing points for SVGP model.
        num_epochs: int, default=20
            Number of epochs (passes through entire dataset) to train SVGP for.
        cv_folds: int, default=1
            How many folds of cross-validation to perform. If 1, there is no cross-validation.
        """
        # get proband and control masks
        ctr_mask, _ = self._get_masks()

        # get matrix of confounds
        conf_mat = self._get_conf_mat()

        # get score        
        score = self._get_score()
        
        if self._use_approx(method=method):
            self.loss = self._svgp_normative_model(conf_mat,score,ctr_mask,nu=nu,length_scale=length_scale, length_scale_bounds=length_scale_bounds,
                                                batch_size=batch_size,n_inducing=n_inducing,num_epochs=num_epochs,cv_folds=cv_folds)

        else:
            kernel = ConstantKernel() + WhiteKernel(noise_level=1) + Matern(length_scale=length_scale, nu=nu,length_scale_bounds=length_scale_bounds)
            gp = gaussian_process.GaussianProcessRegressor(kernel=kernel)

            # Define independent and response variables
            y = score[ctr_mask].reshape(-1,1)
            X = conf_mat[ctr_mask]

            if cv_folds == 1:
                gp.fit(X, y)
                y_pred, sigma = gp.predict(conf_mat, return_std=True)
                y_true = self.data[self.score].to_numpy().reshape(-1,1)

                rmse = RMSE(y_true[ctr_mask],y_pred[ctr_mask])
                smse = SMSE(y_true[ctr_mask],y_pred[ctr_mask])
                msll = MSLL(y_true[ctr_mask],y_pred[ctr_mask],sigma[ctr_mask])
            else:
                kf = KFold(n_splits=cv_folds, shuffle=True)
                rmse = []
                smse = []
                msll = []
                print(f'Starting {cv_folds} folds of CV...')
                for i, (train_index, test_index) in enumerate(kf.split(X)):
                    X_train, X_test = X[train_index], X[test_index]
                    y_train, y_test = y[train_index], y[test_index]
                    gp.fit(X_train, y_train)
                    y_pred, sigma = gp.predict(X_test, return_std=True)

                    r = RMSE(y_test,y_pred)
                    s = SMSE(y_test,y_pred)
                    m = MSLL(y_test,y_pred,sigma)
                    print(f'CV Fold {i}: RMSE={r:.3f} - SMSE={s:.3f} - MSLL={m:.3f}')
                    rmse.append(r)
                    smse.append(s)
                    msll.append(m)
                print('Done!')

                rmse = np.mean(rmse)
                smse = np.mean(smse)
                msll = np.mean(msll)
                print(f'Average: RMSE={rmse:.3f} - SMSE={smse:.3f} - MSLL={msll:.3f}')

                gp.fit(X, y)
                y_pred, sigma = gp.predict(conf_mat, return_std=True)
                y_true = self.data[self.score].to_numpy().reshape(-1,1)

            self.data['GP_pred'] = y_pred
            self.data['GP_sigma'] = sigma
            self.data['GP_residuals'] = y_true - y_pred
            self.data['GP_z'] = self.data['GP_residuals'] / self.data['GP_sigma']

            self.RMSE_GP = rmse
            self.SMSE_GP = smse
            self.MSLL_GP = msll

        self._test_gp_residuals(conf_mat)

    def _svgp_normative_model(self,conf_mat,score,ctr_mask,nu=2.5,length_scale=1,length_scale_bounds=(1e-5,1e5),
                                batch_size=256,n_inducing=500,num_epochs=20,cv_folds=1):
        """ Compute SVGP model. See GPyTorch documentation for further details:
        https://docs.gpytorch.ai/en/v1.1.1/examples/04_Variational_and_Approximate_GPs/SVGP_Regression_CUDA.html#Creating-a-SVGP-Model.

        Parameters
        ----------
        conf_mat: array
            Confounds with categorical values dummy encoded.
        score: array
            Score/response variable.
        ctr_mask: array
            Mask (boolean array) with controls marked True.
        length_scale: float, default=1
            Length scale parameter of Matern kernel.
        length_scale_bounds: pair of floats >= 0 or 'fixed', default=(1e-5, 1e5)
            The lower and upper bound on length_scale. If set to 'fixed', ‘length_scale’ cannot be changed during hyperparameter tuning.
        nu: float, default=2.5
            Nu parameter of Matern kernel.
        batch_size: int, default=256
            Batch size for SVGP model training and prediction.
        n_inducing: int, default=500
            Number of inducing points for SVGP model.
        num_epochs: int, default=20
            Number of epochs (passes through entire dataset) to train SVGP for.
        cv_folds: int, default=1
            How many folds of cross-validation to perform. If 1, there is no cross-validation.
        
        Raises
        ------
        ImportError
            GPyTorch or it's dependencies aren't installed.

        Returns
        -------
        array
            Loss per epoch of training (temp for debugging).
        """
        try:
            from pynm.models.approx import SVGP
        except:
            raise ImportError("GPyTorch and it's dependencies must be installed to use the SVGP model.")
        else:
            if cv_folds == 1:
                svgp = SVGP(conf_mat[ctr_mask],conf_mat,score[ctr_mask],score,n_inducing=n_inducing,batch_size=batch_size,nu=nu,
                            length_scale=length_scale,length_scale_bounds=length_scale_bounds)
                
                svgp.train(num_epochs=num_epochs)
                means, sigma = svgp.predict()

                y_pred = means.numpy()
                y_true = score
                residuals = (y_true - y_pred).astype(float)

                rmse = RMSE(y_true[ctr_mask],y_pred[ctr_mask])
                smse = SMSE(y_true[ctr_mask],y_pred[ctr_mask])
                msll = MSLL(y_true[ctr_mask],y_pred[ctr_mask],sigma.numpy()[ctr_mask])

            else:
                X = conf_mat[ctr_mask]
                y = score[ctr_mask]

                kf = KFold(n_splits=cv_folds, shuffle=True)
                rmse = []
                smse = []
                msll = []
                print(f'Starting {cv_folds} folds of CV...')
                for i, (train_index, test_index) in enumerate(kf.split(X)):
                    X_train, X_test = X[train_index], X[test_index]
                    y_train, y_test = y[train_index], y[test_index]

                    cv_svgp = SVGP(X_train,X_test,y_train,y_test,n_inducing=n_inducing,batch_size=batch_size,nu=nu,
                            length_scale=length_scale,length_scale_bounds=length_scale_bounds)
                
                    cv_svgp.train(num_epochs=num_epochs)
                    cv_means, cv_sigma = cv_svgp.predict()

                    cv_y_pred = cv_means.numpy()
                    cv_residuals = (y_test - cv_y_pred).astype(float)

                    r = RMSE(y_test,cv_y_pred)
                    s = SMSE(y_test,cv_y_pred)
                    m = MSLL(y_test,cv_y_pred,cv_sigma.numpy())

                    print(f'CV Fold {i}: RMSE={r:.3f} - SMSE={s:.3f} - MSLL={m:.3f}')
                    rmse.append(r)
                    smse.append(s)
                    msll.append(m)
                print('Done!')

                rmse = np.mean(rmse)
                smse = np.mean(smse)
                msll = np.mean(msll)
                print(f'Average: RMSE={rmse:.3f} - SMSE={smse:.3f} - MSLL={msll:.3f}')

                svgp = SVGP(conf_mat[ctr_mask],conf_mat,score[ctr_mask],score,n_inducing=n_inducing,batch_size=batch_size,nu=nu,
                            length_scale=length_scale,length_scale_bounds=length_scale_bounds)
                
                svgp.train(num_epochs=num_epochs)
                means, sigma = svgp.predict()

                y_pred = means.numpy()
                y_true = score
                residuals = (y_true - y_pred).astype(float)

            self.data['GP_pred'] = y_pred
            self.data['GP_sigma'] = sigma.numpy()
            self.data['GP_residuals'] = residuals
            self.data['GP_z'] = self.data['GP_residuals']/self.data['GP_sigma']

            self.RMSE_GP = rmse
            self.SMSE_GP = smse
            self.MSLL_GP = msll

    
    def gamlss_normative_model(self,mu=None,sigma=None,nu=None,tau=None,family='SHASHo2',method='RS',cv_folds=1):
        """Compute GAMLSS normative model.
        
        Parameters
        ----------
        mu: str or None
            Formula for mu (location) parameter of GAMLSS. If None, formula for score is sum of confounds
            with non-categorical columns as smooth functions, e.g. "score ~ ps(age) + sex".
        sigma: str or None
            Formula for sigma (scale) parameter of GAMLSS. If None, formula is '~ 1'.
        nu: str or None
            Formula for nu (skewness) parameter of GAMLSS. If None, formula is '~ 1'.
        tau: str or None
            Formula for tau (kurtosis) parameter of GAMLSS. If None, formula is '~ 1'.
        family: str,default='SHASHo2'
            Family of distributions to use for fitting, default is 'SHASHo2'. See R documentation for GAMLSS package for other available families of distributions.
        method: str, default = 'RS'
            Method for fitting GAMLSS. Can be 'RS' (Rigby and Stasinopoulos algorithm), 'CG' (Cole and Green algorithm) or 'mixed(n,m)' where n & m are integers.
            Specifying 'mixed(n,m)' will use the RS algorithm for n iterations and the CG algorithm for up to m additional iterations.
        cv_folds: int, default=1
            How many folds of cross-validation to perform. If 1, there is no cross-validation.
        
        Notes
        -----
        If using 'random()' to model a random effect in any of the formulas, it must be passed a column of the dataframe with categorical values
        as a factor: e.g. 'random(as.factor(COL))'.
        """
        try:
            from pynm.models.gamlss import GAMLSS
        except:
            raise ImportError("R and the GAMLSS package must be installed to use GAMLSS model, see documentation for installation help.")
        else:
            # get proband and control masks
            ctr_mask, _ = self._get_masks()

            gamlss = GAMLSS(mu=mu,sigma=sigma,nu=nu,tau=tau,family=family,method=method,
                            score=self.score,confounds=self.confounds)

            nan_cols = ['LOESS_pred','LOESS_residuals','LOESS_z','LOESS_rank','LOESS_sigma',
            'Centiles_pred','Centiles_residuals','Centiles_z','Centiles','Centiles_rank','Centiles_sigma',
            'Centiles_95','Centiles_5','Centiles_32','Centiles_68']
            gamlss_data = self.data[[c for c in self.data.columns if c not in nan_cols]]

            if cv_folds == 1:
                gamlss.fit(gamlss_data[ctr_mask])
                
                mu_pred = gamlss.predict(gamlss_data,what='mu')
                sigma_pred = gamlss.predict(gamlss_data,what='sigma')
                
                rmse = RMSE(self.data[self.score].values[ctr_mask],mu_pred[ctr_mask])
                smse = SMSE(self.data[self.score].values[ctr_mask],mu_pred[ctr_mask])
                msll = MSLL(self.data[self.score].values[ctr_mask],mu_pred[ctr_mask],sigma_pred[ctr_mask])
            
            else:
                X = gamlss_data[ctr_mask]
                kf = KFold(n_splits=cv_folds, shuffle=True)
                rmse = []
                smse = []
                msll = []
                print(f'Starting {cv_folds} folds of CV...')
                for i, (train_index, test_index) in enumerate(kf.split(X)):
                    X_train, X_test = X.iloc[train_index], X.iloc[test_index]

                    gamlss.fit(X_train)

                    cv_mu_pred = gamlss.predict(X_test,what='mu')
                    cv_sigma_pred = gamlss.predict(X_test,what='sigma')

                    r = RMSE(X_test[self.score].values,cv_mu_pred)
                    s = SMSE(X_test[self.score].values,cv_mu_pred)
                    m = MSLL(X_test[self.score].values,cv_mu_pred,cv_sigma_pred)
                    print(f'CV Fold {i}: RMSE={r:.3f} - SMSE={s:.3f} - MSLL={m:.3f}')
                    rmse.append(r)
                    smse.append(s)
                    msll.append(m)
                print('Done!')

                rmse = np.mean(rmse)
                smse = np.mean(smse)
                msll = np.mean(msll)
                print(f'Average: RMSE={rmse:.3f} - SMSE={smse:.3f} - MSLL={msll:.3f}')

                gamlss.fit(gamlss_data[ctr_mask])
                
                mu_pred = gamlss.predict(gamlss_data,what='mu')
                sigma_pred = gamlss.predict(gamlss_data,what='sigma')

            self.data['GAMLSS_pred'] = mu_pred
            self.data['GAMLSS_sigma'] = sigma_pred
            self.data['GAMLSS_residuals'] = self.data[self.score] - self.data['GAMLSS_pred']
            self.data['GAMLSS_z'] = self.data['GAMLSS_residuals']/self.data['GAMLSS_sigma']

            self.RMSE_GAMLSS = rmse
            self.SMSE_GAMLSS = smse
            self.MSLL_GAMLSS = msll
    
    def report(self):
        """ Prints the values of each metric (SMSE, RMSE, MSLL) for the models that have been run.
        """
        print("------\nReport\n------")
        models = []
        for m in ['LOESS','Centiles','GP','GAMLSS']:
            if '{}_pred'.format(m) in self.data.columns:
                models.append(m)
        if len(models)==0:
            print('No models have been run.')
            return
        
        print("========= SMSE - RMSE - MSLL")
        for m in models:
            k = 9 - len(m)
            m_formatted = m + k*' '
            smse = np.round(eval(f"self.SMSE_{m}"),2)
            if np.isnan(smse):
                smse = 'NaN '
            rmse = np.round(eval(f"self.RMSE_{m}"),2)
            if np.isnan(rmse):
                rmse = 'NaN '
            msll = 'N/A'
            if (m == 'GP') or (m == 'GAMLSS'):
                msll = np.round(eval(f"self.MSLL_{m}"),2)
            print(f"{m_formatted} {smse}   {rmse}   {msll}")

    def _plot(self, ax,kind=None,gp_xaxis=None,gamlss_xaxis=None):
        """ Plot the data with the normative model overlaid.

        Parameters
        ----------
        ax: matplotlib axis
            Axis on which to plot.
        kind: str, default=None
            Type of plot among "LOESS" (local polynomial), "Centiles", "GP" (gaussian processes), 
            or "GAMLSS" (generalized additive models of location scale and shape). 
        gp_xaxis: str,default=None
            Which confound to use for xaxis of GP plot. If set to None, first confound in list passed to model will be used.
        gamlss_xaxis: str,default=None
            Which confound to use for xaxis of GAMLSS plot. If set to None, first confound in list passed to model will be used.

        Returns
        -------
        Axis
            handle for the matplotlib axis of the plot
        """
        if kind is None:
            sns.scatterplot(data=self.data, x=self.conf, y=self.score,
                             hue=self.group, style=self.group,ax=ax)
        elif kind == 'LOESS':
            sns.scatterplot(data=self.data, x=self.conf, y=self.score,
                             hue=self.group, style=self.group,ax=ax)
            tmp=self.data.sort_values(self.conf)
            ax.plot(tmp[self.conf], tmp['LOESS_pred'], '-k',label='Prediction')
            ax.plot(tmp[self.conf], tmp['LOESS_pred'] - 1.96*tmp['LOESS_sigma'], '--k')
            ax.plot(tmp[self.conf], tmp['LOESS_pred'] + 1.96*tmp['LOESS_sigma'], '--k',label='95% CI')
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels)
            ax.set_title(f"{kind} SMSE={self.SMSE_LOESS:.3f}")
        elif kind == 'Centiles':
            sns.scatterplot(data=self.data, x=self.conf, y=self.score,
                                hue=self.group, style=self.group,ax=ax)
            tmp=self.data.sort_values(self.conf)
            ax.plot(tmp[self.conf], tmp['Centiles_pred'], '-k',label='Prediction')
            ax.plot(tmp[self.conf], tmp['Centiles_5'],'--k')
            ax.plot(tmp[self.conf], tmp['Centiles_95'],'--k',label='95% CI')
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels)
            ax.set_title(f"{kind} SMSE={self.SMSE_Centiles:.3f}")
        elif kind == 'GP':
            if gp_xaxis is None:
                gp_xaxis = self.conf
            sns.scatterplot(data=self.data, x=gp_xaxis, y=self.score,
                                hue=self.group, style=self.group,ax=ax)
            tmp=self.data.sort_values(gp_xaxis)
            if len(self.confounds) == 1:
                ax.plot(tmp[gp_xaxis], tmp['GP_pred'], '-k',label='Prediction')
                ax.plot(tmp[gp_xaxis], tmp['GP_pred'] - 1.96*tmp['GP_sigma'], '--k')
                ax.plot(tmp[gp_xaxis], tmp['GP_pred'] + 1.96*tmp['GP_sigma'], '--k',label='95% CI')
            else:
                ax.scatter(tmp[gp_xaxis], tmp['GP_pred'], label='Prediction',color='black',marker='_',s=25)
                ax.scatter(tmp[gp_xaxis], tmp['GP_pred'] - 1.96*tmp['GP_sigma'],color='black',s=0.2)
                ax.scatter(tmp[gp_xaxis], tmp['GP_pred'] + 1.96*tmp['GP_sigma'], label='95% CI',color='black',s=0.2)
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels)
            ax.set_title(f"{kind} SMSE={self.SMSE_GP:.3f} - MSLL={self.MSLL_GP:.3f}")
        elif kind == 'GAMLSS':
            if gamlss_xaxis is None:
                gamlss_xaxis = self.conf
            sns.scatterplot(data=self.data, x=gamlss_xaxis, y=self.score,
                                hue=self.group, style=self.group,ax=ax)
            tmp=self.data.sort_values(gamlss_xaxis)
            if len(self.confounds) == 1:
                ax.plot(tmp[gamlss_xaxis], tmp['GAMLSS_pred'], '-k',label='Prediction')
                ax.plot(tmp[gamlss_xaxis], tmp['GAMLSS_pred'] - 1.96*tmp['GAMLSS_sigma'], '--k')
                ax.plot(tmp[gamlss_xaxis], tmp['GAMLSS_pred'] + 1.96*tmp['GAMLSS_sigma'], '--k',label='95% CI')
            else:
                ax.scatter(tmp[gamlss_xaxis], tmp['GAMLSS_pred'], label='Prediction',color='black',marker='_',s=25)
                ax.scatter(tmp[gamlss_xaxis], tmp['GAMLSS_pred'] - 1.96*tmp['GAMLSS_sigma'],color='black',s=0.2)
                ax.scatter(tmp[gamlss_xaxis], tmp['GAMLSS_pred'] + 1.96*tmp['GAMLSS_sigma'], label='95% CI',color='black',s=0.2)
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels)
            ax.set_title(f"{kind} SMSE={self.SMSE_GAMLSS:.3f} - MSLL={self.MSLL_GAMLSS:.3f}")
        return ax

    def plot(self, kind=None,gp_xaxis=None,gamlss_xaxis=None):
        """Plot the data with the normative model overlaid.

        Parameters
        ----------
        kind: list, default=None
            Type of plot, must be a valid subset of ["Centiles","LOESS","GP","GAMLSS"] or None. If None, all available
            results will be plotted, if None are available a warning will be raised and only the data will be plotted.
        gp_xaxis: str,default=None
            Which confound to use for xaxis of GP plot. If set to None, first confound in list passed to model will be used.
        gamlss_xaxis: str,default=None
            Which confound to use for xaxis of GAMLSS plot. If set to None, first confound in list passed to model will be used.
        
        Raises
        ------
        ValueError
            Plot kind not recognized, must be a valid subset of ["Centiles","LOESS","GP","GAMLSS"] or None.
        """
        if kind is None:
            kind = []
            for k in ['LOESS','Centiles','GP','GAMLSS']:
                if '{}_pred'.format(k) in self.data.columns:
                    kind.append(k)
            if len(kind)==0:
                warnings.warn('No model results found in data.')
        
        if set(kind).issubset(set(['LOESS','Centiles','GP','GAMLSS'])) and len(kind)>1:
            fig, ax = plt.subplots(1,len(kind),figsize=(len(kind)*5,5))
            for i,k in enumerate(kind):
                self._plot(ax[i],kind=k,gp_xaxis=gp_xaxis,gamlss_xaxis=gamlss_xaxis)
            plt.show()
        elif set(kind).issubset(set(['LOESS','Centiles','GP','GAMLSS'])) and len(kind)>0:
            fig, ax = plt.subplots(1,len(kind),figsize=(len(kind)*5,5))
            self._plot(ax,kind=kind[0],gp_xaxis=gp_xaxis,gamlss_xaxis=gamlss_xaxis)
            plt.show()
        elif len(kind)==0:
            fig, ax = plt.subplots(1,1)
            self._plot(ax,None,gp_xaxis=gp_xaxis,gamlss_xaxis=gamlss_xaxis)
            ax.set_title('Data')
            plt.show()
        else:
            raise ValueError('Plot kind not recognized, must be a valid subset of ["Centiles","LOESS","GP","GAMLSS"] or None.')

    def _plot_res_z(self, ax,kind=None, confound=None,z=False):
        """ Plot the residuals of the normative model.

        Parameters
        ----------
        ax: matplotlib axis
            Axis on which to plot.
        kind: str, default=None
            Type of plot among "LOESS" (local polynomial), "Centiles", "GP" (gaussian processes),
            or "GAMLSS" (generalized additive models of location scale and shape).
        confound: str or None
            Which confound to use as xaxis of plot, must be categorical or None.
        """
        if kind == 'LOESS':
            if z:
                sns.violinplot(x=confound, y='LOESS_z',
                            data=self.data, split=True, palette='Blues', hue=self.group,ax=ax)
            else:
                sns.violinplot(x=confound, y='LOESS_residuals',
                            data=self.data, split=True, palette='Blues', hue=self.group,ax=ax)
            ax.set_title(f"{kind} SMSE={self.SMSE_LOESS:.3f}")
        if kind == 'Centiles':
            if z:
                sns.violinplot(x=confound, y='Centiles_z',
                           data=self.data, split=True, palette='Blues', hue=self.group,ax=ax)
            else:
                sns.violinplot(x=confound, y='Centiles_residuals',
                            data=self.data, split=True, palette='Blues', hue=self.group,ax=ax)
            ax.set_title(f"{kind} SMSE={self.SMSE_Centiles:.3f}")
        if kind == 'GP':
            if z:
                    sns.violinplot(x=confound, y='GP_z',
                            data=self.data, split=True, palette='Blues', hue=self.group,ax=ax)
            else:
                sns.violinplot(x=confound, y='GP_residuals',
                            data=self.data, split=True, palette='Blues', hue=self.group,ax=ax)
            ax.set_title(f"{kind} SMSE={self.SMSE_GP:.3f} - MSLL={self.MSLL_GP:.3f}")
        if kind == 'GAMLSS':
            if z:
                sns.violinplot(x=confound, y='GAMLSS_z',
                            data=self.data, split=True, palette='Blues', hue=self.group,ax=ax)
            else:
                sns.violinplot(x=confound, y='GAMLSS_residuals',
                            data=self.data, split=True, palette='Blues', hue=self.group,ax=ax)
            ax.set_title(f"{kind} SMSE={self.SMSE_GAMLSS:.3f} - MSLL={self.MSLL_GAMLSS:.3f}")
        if not isinstance(confound,str):
            ax.set_xticklabels([''])
    
    def _plot_res_z_cont(self, ax,kind=None, confound=None,z=False):
        """ Plot the residuals of the normative model.

        Parameters
        ----------
        ax: matplotlib axis
            Axis on which to plot.
        kind: str, default=None
            Type of plot among "LOESS" (local polynomial), "Centiles", "GP" (gaussian processes),
            or "GAMLSS" (generalized additive models of location scale and shape).
        confound: str or None
            Which confound to use as xaxis of plot, must be continuous.
        """
        if kind == 'LOESS':
            if z:
                sns.scatterplot(x=confound, y='LOESS_z',
                                data=self.data, hue=self.group,ax=ax)
            else:
                sns.scatterplot(x=confound, y='LOESS_residuals',
                                data=self.data, hue=self.group,ax=ax)
            ax.set_title(f"{kind} SMSE={self.SMSE_LOESS:.3f}")
        if kind == 'Centiles':
            if z:
                sns.scatterplot(x=confound, y='Centiles_z',
                            data=self.data, hue=self.group,ax=ax)
            else:
                sns.scatterplot(x=confound, y='Centiles_residuals',
                                data=self.data, hue=self.group,ax=ax)
            ax.set_title(f"{kind} SMSE={self.SMSE_Centiles:.3f}")
        if kind == 'GP':
            if z:
                sns.scatterplot(x=confound, y='GP_z',
                                data=self.data, hue=self.group,ax=ax)
            else:
                sns.scatterplot(x=confound, y='GP_residuals',
                                data=self.data, hue=self.group,ax=ax)
            ax.set_title(f"{kind} SMSE={self.SMSE_GP:.3f} - MSLL={self.MSLL_GP:.3f}")
        if kind == 'GAMLSS':
            if z:
                sns.scatterplot(x=confound, y='GAMLSS_z',
                                data=self.data, hue=self.group,ax=ax)
            else:
                sns.scatterplot(x=confound, y='GAMLSS_residuals',
                                data=self.data, hue=self.group,ax=ax)
            ax.set_title(f"{kind} SMSE={self.SMSE_GAMLSS:.3f} - MSLL={self.MSLL_GAMLSS:.3f}")

    def plot_res(self, kind=None, confound=None):
        """Plot the residuals of the normative model.

        Parameters
        ----------
        kind: list default=None
            Type of plot, must be a valid subset of ["Centiles","LOESS","GP","GAMLSS"] or None. If None, all available
            results will be plotted, if None are available a ValueError will be raised.
        confound: str, default=None
            Which confound to use as xaxis of plot.
        
        Raises
        ------
        ValueError
            Plot kind not recognized, must be a valid subset of ["Centiles","LOESS","GP","GAMLSS"] or None.
        ValueError
            No model results found in data.
        """
        _, cat = read_confounds(self.confounds)
        if confound is None: 
            categorical = True
        elif confound in cat: 
            categorical = True
        else: 
            categorical = False

        if kind is None:
            kind = []
            for k in ['LOESS','Centiles','GP','GAMLSS']:
                if '{}_residuals'.format(k) in self.data.columns:
                    kind.append(k)
            if len(kind)==0:
                raise ValueError('No model residuals found in data.')
        
        if set(kind).issubset(set(['LOESS','Centiles','GP','GAMLSS'])) and len(kind)>1:
            fig, ax = plt.subplots(1,len(kind),figsize=(len(kind)*5,5))
            for i,k in enumerate(kind):
                if categorical:
                    self._plot_res_z(ax[i],kind=k,confound=confound)
                else:
                    self._plot_res_z_cont(ax[i],kind=k,confound=confound)
            plt.show()
        elif set(kind).issubset(set(['LOESS','Centiles','GP','GAMLSS'])):
            fig, ax = plt.subplots(1,len(kind),figsize=(len(kind)*5,5))
            if categorical:
                self._plot_res_z(ax,kind=kind[0],confound=confound)
            else:
                self._plot_res_z_cont(ax,kind=kind[0],confound=confound)
            plt.show()
        else:
            raise ValueError('Plot kind not recognized, must be a valid subset of ["Centiles","LOESS","GP","GAMLSS"] or None.')

    def plot_z(self, kind=None, confound=None):
        """Plot the deviance scores of the normative model.

        Parameters
        ----------
        kind: list default=None
            Type of plot, must be a valid subset of ["Centiles","LOESS","GP","GAMLSS"] or None. If None, all available
            results will be plotted, if None are available a ValueError will be raised.
        confound: str, default=None
            Which confound to use as xaxis of plot.
        
        Raises
        ------
        ValueError
            Plot kind not recognized, must be a valid subset of ["Centiles","LOESS","GP","GAMLSS"] or None.
        ValueError
            No model results found in data.
        """
        _, cat = read_confounds(self.confounds)
        if confound is None: 
            categorical = True
        elif confound in cat: 
            categorical = True
        else: 
            categorical = False

        if kind is None:
            kind = []
            for k in ['LOESS','Centiles','GP','GAMLSS']:
                if '{}_z'.format(k) in self.data.columns:
                    kind.append(k)
            if len(kind)==0:
                raise ValueError('No model deviance scores found in data.')
        
        if set(kind).issubset(set(['LOESS','Centiles','GP','GAMLSS'])) and len(kind)>1:
            fig, ax = plt.subplots(1,len(kind),figsize=(len(kind)*5,5))
            for i,k in enumerate(kind):
                if categorical:
                    self._plot_res_z(ax[i],kind=k,confound=confound,z=True)
                else:
                    self._plot_res_z_cont(ax[i],kind=k,confound=confound,z=True)
            plt.show()
        elif set(kind).issubset(set(['LOESS','Centiles','GP','GAMLSS'])):
            fig, ax = plt.subplots(1,len(kind),figsize=(len(kind)*5,5))
            if categorical:
                self._plot_res_z(ax,kind=kind[0],confound=confound,z=True)
            else:
                self._plot_res_z_cont(ax,kind=kind[0],confound=confound,z=True)
            plt.show()
        else:
            raise ValueError('Plot kind not recognized, must be a valid subset of ["Centiles","LOESS","GP","GAMLSS"] or None.')



import re
import rpy2.robjects as ro
from rpy2.robjects.packages import importr
from rpy2.robjects import numpy2ri
from rpy2.robjects import pandas2ri
from rpy2.robjects import r
from pynm.util import read_confounds

class GAMLSS:
    """Class for GAMLSS model.
        
        Attributes
        ----------
        gamlss_data: R package
            Python imported R package.
        gamlss_dist: R package
            Python imported R package.
        gamlss: R package
            Python imported R package.
        mu_f: R formula
            Formula for mu (location) parameter.
        sigma_f: R formula
            Formula for sigma (scale) parameter.
        nu_f: R formula
            Formula for nu (skewness) parameter.
        tau_f: R formula
            Formula for tau (kurtosis) parameter.
        rfamily: R object
            Family of distributions to use for fitting.
        method: str
            Method to fit GAMLSS.
        model: R object
            Fitted GAMLSS model.
        """

    def __init__(self,mu=None,sigma=None,nu=None,tau=None,family='SHASHo2',method='RS',score=None,confounds=None):
        """Create GAMLSS object. Formulas must be written for R, using functions available in the GAMLSS package.
        
        Parameters
        ----------
        mu: str, default=None
            Formula for mu (location) parameter of GAMLSS. If None, formula for score is sum of confounds
            with non-categorical columns as smooth functions, e.g. "score ~ ps(age) + sex".
        sigma: str, default=None
            Formula for sigma (scale) parameter of GAMLSS. If None, formula is '~ 1'.
        nu: str, default=None
            Formula for nu (skewness) parameter of GAMLSS. If None, formula is '~ 1'.
        tau: str, default=None
            Formula for tau (kurtosis) parameter of GAMLSS. If None, formula is '~ 1'.
        family: str,default='SHASHo2'
            Family of distributions to use for fitting, default is 'SHASHo2'. See R documentation for GAMLSS package for other available families of distributions.
        method: str, default = 'RS'
            Method for fitting GAMLSS. Can be 'RS' (Rigby and Stasinopoulos algorithm), 'CG' (Cole and Green algorithm) or 'mixed(n,m)' where n & m are integers.
            Specifying 'mixed(n,m)' will use the RS algorithm for n iterations and the CG algorithm for up to m additional iterations.
        score: str, default=None
            Label of score in DataFrame.
        confounds: list, default=None
            List of labels of confounds in DataFrame.
        
        Notes
        -----
        If using 'random()' to model a random effect in any of the formulas, it must be passed a column of the dataframe with categorical values
        as a factor: e.g. 'random(as.factor(COL))'.
        """
        numpy2ri.activate()
        pandas2ri.activate()

        self.gamlss_data = importr('gamlss.data')
        self.gamlss_dist = importr('gamlss.dist')
        self.gamlss = importr('gamlss')
        self.base = importr('base')
        
        self.score = score
        self.confounds = confounds
        self.mu_f,self.sigma_f,self.nu_f,self.tau_f = self._get_r_formulas(mu,sigma,nu,tau)
        self.family = family
        self.method = self._get_method(method)
        try:
            self.rfamily = r[family]
        except:
            raise ValueError("Provided family not valid, choose 'SHASHo2', 'NO' or see R documentation for GAMLSS package for other available families of distributions.")
    
    def _get_r_formulas(self,mu,sigma,nu,tau):
        """Convert from string input to R formula.

        Parameters
        ----------
        mu: str or None
            Formula for mu (location) parameter of GAMLSS. If None, formula for score is sum of confounds
            with non-categorical columns as smooth functions, e.g. "score ~ ps(age) + sex".
        sigma: str or None
            Formula for sigma (scale) parameter of GAMLSS. If None, formula is '~ 1'.
        nu: str or None
            Formula for nu (skewness) parameter of GAMLSS. If None, formula is '~ 1'.
        tau: str or None
            Formula for tau (kurtosis) parameter of GAMLSS. If None, formula is '~ 1'.

        Raises
        ------
        ValueError
            If any of the input strings contains a function call not recognised by the R GAMLSS package.
        ValueError
            If mu is None and either score or confounds is None.
        
        Returns
        -------
        R formula, R formula, R formula, R formula
            R formula equivalent for each input string.
        """
        if mu is None:
            if (self.score is None) or (self.confounds is None):
                raise ValueError('If mu is None, both score and confounds must be provided i.e. not None.')
            _,cat = read_confounds(self.confounds)
            formula_conf = ['ps({})'.format(conf) for conf in self.confounds if not conf[2:-1] in cat] + cat
            mu = '{} ~ {}'.format(self.score,' + '.join(formula_conf))
        if sigma is None:
            sigma = '~ 1'
        if nu is None:
            nu = '~ 1'
        if tau is None:
            tau = '~ 1'
        
        # get r functions from formulas
        p = re.compile(r"\w*\(")
        funcs = []
        for s in [mu,sigma,nu,tau]:
            for f in p.findall(s):
                funcs.append(f[:-1])

        for func in funcs:
            try:
                exec("{} = r['{}']".format(func,func))
            except:
                raise ValueError("'{}' function not found in R GAMLSS package. See GAMLSS documentation for available functions.".format(func))

        return mu,sigma,nu,tau
    
    def _get_method(self,method):
        """ Get method parameter in appropriate format for R.

        Raises
        ------
        TypeError
            "Argument 'method' must be of type str."
        ValueError
            "Unrecognized argument for 'method'."
        """
        if not isinstance(method,str):
            raise TypeError("Argument 'method' must be of type str.")

        pattern = re.compile(r"mixed\([0-9]*,[0-9]*\)")
    
        if method == 'RS':
            return 'RS()'
        elif method == 'CG':
            return 'CG()'
        elif pattern.match(method) is not None:
            return method
        else:
            raise ValueError("Unrecognized argument for 'method'.")
    
    def fit(self,train_data):
        """Create and fit gamlss model.

        Parameters
        ----------
        train_data: DataFrame
            DataFrame with training data.
        """
        ro.globalenv['train_data'] = train_data

        self.model = r(f'''gamlss({self.mu_f},
                                sigma.formula={self.sigma_f},
                                nu.formula={self.nu_f},
                                tau.formula={self.tau_f},
                                family={self.family},
                                data=train_data,
                                method={self.method})''')
                    
    def predict(self,test_data,what='mu'):
        """Predict from fitted gamlss model.
        
        Parameters
        ----------
        test_data: DataFrame
            DataFrame with test data.
        what: str
            Which parameter to predict, can be 'mu','sigma', 'nu', or 'tau'.
        """
        ro.globalenv['model'] = self.model
        ro.globalenv['test_data'] = test_data

        res = r(f'''predict(model,newdata=test_data,parameter="{what}")''')
        return res



import numpy as np
from scipy.stats.mstats import mquantiles

def centiles_fit(train_data,bins,bin_width):
    """ Fit Centiles model.

    Parameters
    ----------
    train_data: array
        Training data for Centiles model.
    bins: array
        Bins for Centiles model.
    bin_width: float
        Width of each bin.

    Returns
    -------
    array
        Centiles for each bin.
    """
    z = np.zeros([bins.shape[0], 101])  # centiles

    for i, bin_center in enumerate(bins):
        mu = np.array(bin_center)  # bin_center value (age or conf)
        bin_mask = (abs(train_data[:, :1] - mu) <
                        bin_width) * 1.  # one hot mask
        idx = [u for (u, v) in np.argwhere(bin_mask)]
        scores = train_data[idx, 1]

        # if more than 2 non NaN values do the model
        if (~np.isnan(scores)).sum() > 2:
            # centiles
            z[i, :] = mquantiles(scores, prob=np.linspace(0, 1, 101), alphap=0.4, betap=0.4)
        else:
            z[i] = np.nan
    
    return z

def centiles_predict(test_data,bins,z):
    """ Predict from Centiles model.

    Parameters
    ----------
    test_data: array
        Test data for Centiles model. Column 0 is confound, 1 is score.
    bins: array
        Bins for Centiles model.
    z: array
        Centiles for each bin.

    Returns
    -------
    array
        Centile within which each subject falls.
    array
        Centiles for each subject.
    """
    dists = [np.abs(conf - bins) for conf in test_data[:,0]]
    idx = [np.argmin(d) for d in dists]
    centiles = np.array([z[i] for i in idx])

    result = np.zeros(centiles.shape[0])
    max_mask = test_data[:,1] >= np.max(centiles, axis=1)
    min_mask = test_data[:,1] < np.min(centiles, axis=1)
    else_mask = ~(max_mask | min_mask)
    result[max_mask] = 100
    result[min_mask] = 0
    result[else_mask] = np.array([np.argmin(test_data[:,1][i] >= centiles[i]) for i in range(test_data.shape[0])])[else_mask]

    return result, centiles


import math
import torch
import gpytorch
import numpy as np
import statsmodels.api as sm
from tqdm import tqdm
from torch.utils.data import TensorDataset, DataLoader
from gpytorch.models import ApproximateGP
from gpytorch.variational import CholeskyVariationalDistribution
from gpytorch.variational import VariationalStrategy

class GPModel(ApproximateGP):
    """ Class for GPyTorch model.

    Attributes
    ----------
    mean_module : gpytorch Mean
        Module to calculate mean.
    covar_module : gpytorch Kernel
        Module to calculate covariance.
    """
    def __init__(self, inducing_points,nu=2.5,length_scale=1,length_scale_bounds=(1e-5,1e5)):
        """ Create a GPModel object.

        Parameters
        ----------
        inducing_points: array
            Array of inducing points.
        length_scale: float, default=1
            Length scale parameter of Matern kernel.
        length_scale_bounds: pair of floats >= 0 or 'fixed', default=(1e-5, 1e5)
            The lower and upper bound on length_scale. If set to 'fixed', ‘length_scale’ cannot be changed during hyperparameter tuning.
        nu: float, default=2.5
            Nu parameter of Matern kernel.
        
        Raises
        ------
        ValueError
            Invalid argument for length_scale_bounds
        """
        variational_distribution = CholeskyVariationalDistribution(inducing_points.size(0))
        variational_strategy = VariationalStrategy(self, inducing_points, variational_distribution, learn_inducing_locations=True)
        super(GPModel, self).__init__(variational_strategy)

        self.mean_module = gpytorch.means.ConstantMean()

        if length_scale_bounds == 'fixed':
            constraint = gpytorch.constraints.Interval(length_scale - 0.001,length_scale + 0.0001)
        elif isinstance(length_scale_bounds,tuple):
            constraint = gpytorch.constraints.Interval(length_scale_bounds[0],length_scale_bounds[1])
        else:
            raise ValueError('Invalid argument for length_scale_bounds.')
        prior = gpytorch.priors.NormalPrior(length_scale,1)
        self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.MaternKernel(nu=nu,lengthscale_prior=prior),lengthscale_contraint = constraint)

    def forward(self, x):
        """ Calculate forward pass of GPModel.

        Parameters
        ----------
        x: Tensor
            Data tensor.
        
        Returns
        -------
        MultivariateNormal object
        """
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class SVGP:
    """ Class for SVGP model.

    Attributes
    ----------
    train_loader: pytorch DataLoader
        DataLoader for training data.
    test_loader: pytorch DataLoader
        DataLoader for test data.
    inducing_points: array
        Subset of training data to use as inducing points.
    n_train: int
        Number of training points.
    n_test: int
        Number of test points.
    model: GPModel
        Instance of GPModel class.
    likelihood: gpytorch Likelihood
        Gaussian likelihood function.
    loss: list
        Loss for each epoch of training.
    """
    def __init__(self,X_train,X_test,y_train,y_test,n_inducing=500,batch_size=256,nu=2.5,length_scale=1,length_scale_bounds=(1e-5,1e5)):
        """ Create a SVGP object.

        Parameters
        ----------
        X_train: array
            Training confounds with categorical values dummy encoded.
        X_test: array
            Test confounds with categorical values dummy encoded.
        y_train: array
            Training score/response variable.
        y_test: array
            Test score/response variable.
        length_scale: float, default=1
            Length scale parameter of Matern kernel.
        length_scale_bounds: pair of floats >= 0 or 'fixed', default=(1e-5, 1e5)
            The lower and upper bound on length_scale. If set to 'fixed', ‘length_scale’ cannot be changed during hyperparameter tuning.
        nu: float, default=2.5
            Nu parameter of Matern kernel.
        batch_size: int, default=256
            Batch size for SVGP model training and prediction.
        n_inducing: int, default=500
            Number of inducing points for SVGP model.
        """
        # Get data in torch format
        train_x = torch.from_numpy(X_train).contiguous()
        test_x = torch.from_numpy(X_test).double().contiguous()
        train_y = torch.from_numpy(y_train).contiguous()
        test_y = torch.from_numpy(y_test).double().contiguous()

        # Create datasets
        train_dataset = TensorDataset(train_x, train_y)
        test_dataset = TensorDataset(test_x, test_y)

        # Create dataloaders
        self.train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
        self.test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)
        inducing_idx = np.random.choice(np.array(range(train_x.shape[0])),size=n_inducing)
        self.inducing_points = train_x[inducing_idx, :]
        self.n_train = train_y.size(0)
        self.n_test = test_y.size(0)

        self.model = GPModel(inducing_points=self.inducing_points,nu=nu,length_scale=length_scale,length_scale_bounds=length_scale_bounds).double()
        self.likelihood = gpytorch.likelihoods.GaussianLikelihood()
        self.likelihood.initialize(noise=torch.std(train_x))

        if torch.cuda.is_available():
            self.model = self.model.cuda()
            self.likelihood = self.likelihood.cuda()
        
        self.loss = []
    
    def train(self,num_epochs=20):
        """ Trains the SVGP model.

        Parameters
        ----------
        num_epochs: int
            Number of epochs (full passes through dataset) to train for.
        """
        self.model.train()
        self.likelihood.train()

        optimizer = torch.optim.Adam([{'params': self.model.parameters()},{'params': self.likelihood.parameters()}], lr=0.01)

        # Loss object. We're using the VariationalELBO
        mll = gpytorch.mlls.VariationalELBO(self.likelihood, self.model, num_data=self.n_train)

        epochs_iter = tqdm(range(num_epochs), desc="Epoch")
        for i in epochs_iter:
        # Within each iteration, we will go over each minibatch of data
            minibatch_iter = tqdm(self.train_loader, desc="Minibatch", leave=False)
            for x_batch, y_batch in minibatch_iter:
                optimizer.zero_grad()
                output = self.model(x_batch)
                loss = -mll(output, y_batch)
                minibatch_iter.set_postfix(loss=loss.item())
                loss.backward()
                optimizer.step()
            self.loss.append(loss.item())
    
    def predict(self):
        """ Predict from SVGP model.

        Returns
        ----------
        array
            Model predictions (mean of predictive distribution).
        array
            Model uncertainty (standard deviation of predictive distribution).
        """
        self.model.eval()
        self.likelihood.eval()

        mean = torch.tensor([0.])
        sigma = torch.tensor([0.])
        with torch.no_grad():
            for x_batch, y_batch in self.test_loader:
                preds = self.likelihood(self.model(x_batch)) # get likelihood variance + posterior GP variance
                mean = torch.cat([mean, preds.mean.cpu()])
                sigma = torch.cat([sigma, torch.sqrt(preds.variance.cpu())])
        mean = mean[1:]
        sigma = sigma[1:]
        return mean, sigma


import numpy as np
import statsmodels.api as sm
from statsmodels.sandbox.regression.predstd import wls_prediction_std

def loess_fit(train_data,bins,bin_width):
    """ Fit LOESS model.

    Parameters
    ----------
    train_data: array
        Training data for LOESS model. Column 0 is confound, 1 is score.
    bins: array
        Bins for LOESS model.
    bin_width: float
        Width of each bin.

    Returns
    -------
    array
        Mean of each bin.
    array
        Standard deviation of each bin.
    array
        Confidence interval of each bin.
    """
    zm = np.zeros(bins.shape[0])  # mean
    zstd = np.zeros(bins.shape[0])  # standard deviation
    zci = np.zeros([bins.shape[0], 2])  # confidence interval

    for i, bin_center in enumerate(bins):
        mu = np.array(bin_center)  # bin_center value (age or conf)
        bin_mask = (abs(train_data[:, :1] - mu) < bin_width) * 1.
        idx = [u for (u, v) in np.argwhere(bin_mask)]

        scores = train_data[idx, 1]
        adj_conf = train_data[idx, 0] - mu  # confound relative to bin center

        # if more than 2 non NaN values do the model
        if (~np.isnan(scores)).sum() > 2:
            mod = sm.WLS(scores, sm.tools.add_constant(adj_conf, 
                                                        has_constant='add'),
                            missing='drop', weight=bin_mask.flatten()[idx],
                            hasconst=True).fit()
            zm[i] = mod.params[0]  # mean

            # std and confidence intervals
            prstd,_,_ = wls_prediction_std(mod, [0, 0])
            zstd[i] = prstd
            zci[i, :] = mod.conf_int()[0, :]  # [iv_l, iv_u]

        else:
            zm[i] = np.nan
            zci[i] = np.nan
            zstd[i] = np.nan
    
    return zm, zstd, zci

def loess_predict(test_data,bins,zm,zstd):
    """ Predict from LOESS model.

    Parameters
    ----------
    test_data: array
        Test data for LOESS model. Column 0 is confound, 1 is score.
    bins: array
        Bins for LOESS model.
    zm: array
        Mean of each bin.
    zstd: array
        Standard deviation of each bin.

    Returns
    -------
    array
        Mean for each subject.
    array
        Standard deviation for each subject.
    """
    dists = [np.abs(conf - bins) for conf in test_data[:,0]]
    idx = [np.argmin(d) for d in dists]
    m = np.array([zm[i] for i in idx])
    std = np.array([zstd[i] for i in idx])

    return m, std


from argparse import ArgumentParser
from pynm import pynm
import pandas as pd

def _cli_parser():
    """Reads command line arguments and returns input specifications

    Returns
    -------
    dict
        Parsed arguments.
    """
    parser = ArgumentParser()
    parser.add_argument("--pheno_p",dest='pheno_p',required=True,
                        help="Path to phenotype data. Data must be in a .csv file.")
    parser.add_argument("--out_p",dest='out_p',required=True,
                        help="Path to output directory.")
    parser.add_argument("--confounds",dest='confounds',required=True,
                        help="List of confounds to use in the GP model."
                            "The list must formatted as a string with commas between confounds, "
                            "each confound must be a column name from the phenotype .csv file. "
                            "For GP model all confounds will be used, for LOESS and Centiles models "
                            "only the first is used. For GAMLSS all confounds are used "
                            "unless formulas are specified. Categorical values must be denoted by c(var) "
                            "('c' must be lower case), e.g. 'c(SEX)' for column name 'SEX'.")
    parser.add_argument("--score",dest='score',required=True,
                        help="Response variable for all models. "
                        "Must be a column title from phenotype .csv file.")
    parser.add_argument("--group",dest='group',required=True,
                        help="Column name from the phenotype .csv file that "
                        "distinguishes probands from controls. The column must be "
                        "encoded with str labels using 'PROB' for probands and 'CTR' for controls "
                        "or with int labels using 1 for probands and 0 for controls.")
    parser.add_argument("--train_sample",default=1,dest='train_sample',
                        help="Which method to use for a training sample, can be a float in (0,1] "
                        "for a percentage of controls or 'manual' to be manually set using a column "
                        "of the DataFrame labelled 'train_sample'.")
    parser.add_argument("--LOESS",dest='LOESS',action='store_true',
                        help="Flag to run LOESS model.")
    parser.add_argument("--centiles",dest='centiles',action='store_true',
                        help="Flag to run Centiles model.")
    parser.add_argument("--bin_spacing",default = -1,dest='bin_spacing',
                        help="Distance between bins for LOESS & centiles models.")
    parser.add_argument("--bin_width",default = -1,dest='bin_width',
                        help="Width of bins for LOESS & centiles models.")
    parser.add_argument("--GP",dest='GP',action='store_true',
                        help="Flag to run Gaussian Process model.")
    parser.add_argument("--gp_method",default = 'auto',dest='gp_method',
                        help="Method to use for the GP model. Can be set to "
                            "'auto','approx' or 'exact'. In 'auto' mode, "
                            "the exact model will be used for datasets smaller "
                            "than 2000 data points. SVGP is used for the approximate model. "
                            "See documentation for details. Default value is 'auto'.")
    parser.add_argument("--gp_num_epochs",default=20, dest='gp_num_epochs',
                        help="Number of training epochs for SVGP model. "
                            "See documentation for details. Default value is 20.")
    parser.add_argument("--gp_n_inducing",default=500,dest='gp_n_inducing',
                        help="Number of inducing points for SVGP model. "
                            "See documentation for details. Default value is 500.")
    parser.add_argument("--gp_batch_size",default=256,dest='gp_batch_size',
                        help="Batch size for training and predicting from SVGP model. "
                            "See documentation for details. Default value is 256.")
    parser.add_argument("--gp_length_scale",default=1,dest='gp_length_scale',
                        help="Length scale of Matern kernel for exact model. "
                            "See documentation for details. Default value is 1.")
    parser.add_argument("--gp_length_scale_bounds",default=(1e-5,1e5),dest='gp_length_scale_bounds', nargs='*',
                        help="The lower and upper bound on length_scale. If set to 'fixed', "
                        "length_scale cannot be changed during hyperparameter tuning. "
                        "See documentation for details. Default value is (1e-5,1e5).")
    parser.add_argument("--gp_nu",default=2.5,dest='nu',
                        help="Nu of Matern kernel for exact and SVGP model. "
                            "See documentation for details. Default value is 2.5.")
    parser.add_argument("--GAMLSS",dest='GAMLSS',action='store_true',
                        help="Flag to run GAMLSS.")
    parser.add_argument("--gamlss_mu",default=None,dest='gamlss_mu',
                        help="Formula for mu (location) parameter of GAMLSS. Default "
                        "formula for score is sum of confounds with non-categorical "
                        "columns as smooth functions, e.g. 'score ~ ps(age) + sex'.")
    parser.add_argument("--gamlss_sigma",default=None,dest='gamlss_sigma',
                        help="Formula for mu (location) parameter of GAMLSS. Default "
                        "formula is '~ 1'.")
    parser.add_argument("--gamlss_nu",default=None,dest='gamlss_nu',
                        help="Formula for mu (location) parameter of GAMLSS. Default "
                        "formula is '~ 1'.")
    parser.add_argument("--gamlss_tau",default=None,dest='gamlss_tau',
                        help="Formula for mu (location) parameter of GAMLSS. Default "
                        "formula is '~ 1'.")
    parser.add_argument("--gamlss_family",default='SHASHo2',dest='gamlss_family',
                        help="Family of distributions to use for fitting, default is 'SHASHo2'. "
                        "See R documentation for GAMLSS package for other available families of distributions.")
    return parser.parse_args()

def get_bounds(bounds):
    """Converts gp_length_scale_bounds parameter to appropriate type.

    Returns
    -------
    pair of floats >= 0 or 'fixed'
        Appropriate argument for PyNM.gp_normative_model.
    
    Raises
    ------
    ValueError
        Unrecognized argument for gp_length_scale_bounds.
    """
    if isinstance(bounds,list):
        if len(bounds)==1 and bounds[0]=='fixed':
            return 'fixed'
        elif len(bounds)==2:
            return (float(bounds[0]),float(bounds[1]))
        else:
            raise ValueError('Unrecognized argument for gp_length_scale_bounds.')
    else:
        return bounds
        
def main():
    params = vars(_cli_parser())
    
    confounds = params['confounds'].split(',')            
    data = pd.read_csv(params['pheno_p'])
    
    m = pynm.PyNM(data,params['score'],params['group'],params['conf'],confounds,params['train_sample'],
                bin_spacing=params['bin_spacing'], bin_width=params['bin_width'])
    
    #Run models
    if params['LOESS']:
        m.loess_normative_model()
        m.bins_num()
    if params['centiles']:
        m.centiles_normative_model()
        m.bins_num()
    if params['GP']:   
        gp_length_scale_bounds = get_bounds(params['gp_length_scale_bounds'])
        print(gp_length_scale_bounds)
        print(type(gp_length_scale_bounds))
        m.gp_normative_model(length_scale=params['gp_length_scale'],
                        length_scale_bounds = gp_length_scale_bounds, nu=params['gp_nu'], 
                        method=params['gp_method'],batch_size=params['gp_batch_size'],
                        n_inducing=params['gp_n_inducing'],num_epochs=params['gp_num_epochs'])
    if args.GAMLSS:
        m.gamlss_normative_model(mu=params['gamlss_mu'],sigma=params['gamlss_sigma'],nu=params['gamlss_nu'],
                        tau=params['gamlss_tau'],family=params['gamlss_family'])
    
    m.data.to_csv(params['out_p'],index=False)
    
if __name__ == "__main__":
    raise RuntimeError("`pynm/cli.py` should not be run directly. Please install `pynm`.")


import os
from setuptools import setup

setup(
    name="pynm",
    version="1.0.0b8",
    author="Annabelle HARVEY, Guillaume DUMAS",
    author_email="annabelle.harvey@umontreal.ca, guillaume.dumas@ppsp.team",
    description=("Python implementation of Normative Modelling", 
                 "with GAMLSS, Gaussian Processes, LOESS & Centiles approaches."),
    long_description_content_type="text/x-rst",
    license="BSD",
    keywords="gaussian processes statistics modeling",
    url="https://github.com/ppsp-team/PyNM",
    packages=['pynm', 'test', 'pynm/models'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: BSD License",
    ],
    entry_points={
        'console_scripts': [
            'pynm = pynm.cli:main',
        ],
    },
   install_requires=[
        'matplotlib',
        'seaborn',
        'tqdm',
        'numpy >= 1.21.0',
        'pandas >= 1.1.5',
        'scikit_learn >= 0.24.1',
        'scipy >= 1.5.3',
        'statsmodels >= 0.13.2',
        'rpy2 >= 3.1.0',
        'torch >= 1.8.0',
        'gpytorch >= 1.4.0',
   ],
)



from pynm import pynm
import numpy as np
import pandas as pd
import scipy.stats as sp
import math
import pytest
from pynm.util import *
import matplotlib.pyplot as plt
from unittest.mock import patch 
from sklearn.model_selection import train_test_split

def model(age, sex, offset):
    noise = np.random.normal(0, 0.1)
    return 0.001*age-0.00001*(age-50)**2+0.5 + noise - np.random.uniform(0, 0.3) * sex + offset


def model_prob(age, sex, offset):
    noise = np.random.normal(0, 0.1)
    return 0.001*age-0.00001*(age-50)**2+0.5 + noise - np.random.uniform(0, 0.3) * sex - 0.2 * np.random.uniform() + offset

# randseed = 3, sample_size = 1, n_sites = 2 has ONE PROB n=6
# randseed = 1, sample_size = 1, n_sites = 2 has NO PROB n=12
def generate_data(group='PROB_CON', sample_size=1, n_sites=2, randseed=3):
    np.random.seed(randseed)
    n_sites = n_sites
    age_min = (np.random.rand(n_sites)*50).astype(int)
    sites = pd.DataFrame(data={'sex_ratio': np.random.rand(n_sites),
                               'prob_ratio': 0.5*np.random.rand(n_sites),
                               'age_min': age_min,
                               'age_max': (age_min+5+np.random.rand(n_sites)*50).astype(int),
                               'score_shift': np.random.randn(n_sites)/4,
                               'sample_size': (sample_size+np.random.rand(n_sites)*sample_size*10).astype(int)})

    participants = []
    for site in sites.iterrows():
        for participant in range(int(site[1]['sample_size'])):
            sex = np.random.binomial(1, site[1]['sex_ratio'])
            prob = np.random.binomial(1, site[1]['prob_ratio'])
            age = np.random.uniform(site[1]['age_min'], site[1]['age_max'])
            if prob:
                score = model_prob(age, sex, site[1]['score_shift'])
            else:
                score = model(age, sex, site[1]['score_shift'])
            participants.append([site[0], sex, prob, age, score])

    df = pd.DataFrame(participants, columns=['site', 'sex', 'group', 'age', 'score'])
    df.sex.replace({1: 'Female', 0: 'Male'}, inplace=True)
    if group == 'PROB_CON':
        df.group.replace({1: 'PROB', 0: 'CTR'}, inplace=True)
    return df

def sample_x(low=1,high=100,n_subs=1000,sampling='full'):
    if sampling =='full':
        x = np.random.uniform(low=low,high=high,size=n_subs)
    else:
        x = np.concatenate([np.random.normal(20,10,size=int(n_subs/2)),np.random.normal(80,10,size=int(n_subs/2))])
        x = x[(x<high) & (x > low)]
    return x

# Homoskedastic, gaussian noise
def dataset_homo(low=1,high=100,n_subs=1000,sampling='full'):
    x = sample_x(low=low,high=high,n_subs=n_subs,sampling=sampling)
    scores = np.array([np.log(i) + np.random.randn() for i in x])
    df = pd.DataFrame([x,scores],index=['x','score']).transpose()
    df['train_sample'] = 1
    return df

# Homoskedastic, skew noise
def dataset_skew(low=1,high=100,n_subs=1000,sampling='full'):
    x = sample_x(low=low,high=high,n_subs=n_subs,sampling=sampling)
    scores = np.array([np.log(i) + sp.skewnorm.rvs(a=2,size=1)[0] for i in x])
    df = pd.DataFrame([x,scores],index=['x','score']).transpose()
    df['train_sample'] = 1
    return df

# Heteroskedastic linear
def dataset_het(low=1,high=100,n_subs=1000,sampling='full'):
    x = sample_x(low=low,high=high,n_subs=n_subs,sampling=sampling)
    scores = np.array([np.log(i) + 0.15*np.log(i)*np.random.randn() for i in x])
    df = pd.DataFrame([x,scores],index=['x','score']).transpose()
    df['train_sample'] = 1
    return df

class TestBasic:
    def test_read_confounds_some_categorical(self):
        conf = ['a', 'b', 'c(c)']
        clean, cat = read_confounds(conf)
        assert clean == ['a', 'b', 'c']
        assert cat == ['c']

    def test_read_confounds_no_categorical(self):
        conf = ['a', 'b', 'c']
        clean, cat = read_confounds(conf)
        assert clean == conf
        assert cat == []

    def test_read_confounds_all_categorical(self):
        conf = ['c(a)', 'c(b)', 'c(c)']
        clean, cat = read_confounds(conf)
        assert clean == ['a', 'b', 'c']
        assert cat == ['a', 'b', 'c']
    
    def test_invalid_init(self):
        data1 = generate_data(randseed=1)
        data2 = generate_data(randseed=2)
        data = pd.concat([data1,data2])
        assert data.index.nunique() != data.shape[0]
        with pytest.raises(ValueError):
            m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])

    def test_set_group_names_PROB_CON_all_CON(self):
        data = generate_data(randseed=1)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        assert m.CTR == 'CTR'
        assert m.PROB == 'PROB'

    def test_set_group_names_PROB_CON(self):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        assert m.CTR == 'CTR'
        assert m.PROB == 'PROB'

    def test_set_group_names_01(self):
        data = generate_data(randseed=3, group='01')
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        assert m.CTR == 0
        assert m.PROB == 1

    def test_set_group_controls(self):
        data = generate_data(randseed=3, group='01')
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],train_sample=1)
        assert m.group == 'group'
    
    def test_set_group_33(self):
        data = generate_data(randseed=3, group='01')
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],train_sample='0.33')
        assert m.group == 'train_sample'
        assert m.data['train_sample'].sum() == 1
        assert m.data[(m.data['train_sample']==1) & (m.data['group']== 1)].shape[0] == 0
    
    def test_set_group_manual_no_col(self):
        data = generate_data(randseed=3, group='01')
        with pytest.raises(ValueError):
            m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],train_sample='manual')
    
    def test_set_group_manual_zero_col(self):
        data = generate_data(randseed=3, group='01')
        data['train_sample'] = 0
        with pytest.raises(ValueError):
            m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],train_sample='manual')
    
    def test_set_group_manual_good_col(self):
        data = generate_data(randseed=3, group='01')
        data['train_sample'] = [1,1,0,0,0,0]
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],train_sample='manual')
        assert m.PROB == 0
        assert m.group == 'train_sample'

    def test_create_bins(self):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=8,bin_width=1.5)
        m.centiles_normative_model()
        assert m.bins is not None

    def test_bins_num(self):
        data = generate_data(randseed=11)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=5, bin_width=10)
        m._create_bins()
        assert len(m.bins) == 6

    def test_loess_rank(self):
        data = generate_data(randseed=11)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=8,bin_width=1.5)
        m.loess_normative_model()
        assert np.sum(m.data.LOESS_rank) == 1

    def test_loess_normative_model(self):
        data = generate_data(randseed=11)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=8,bin_width=1.5)
        m.loess_normative_model()
        assert math.isclose(2.3482, np.sum(m.data.LOESS_z), abs_tol=0.00001)

    def test_centiles_rank(self):
        data = generate_data(randseed=11)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=8,bin_width=1.5)
        m.centiles_normative_model()
        assert np.sum(m.data.Centiles_rank) == -22

    def test_centiles_normative_model(self):
        data = generate_data(randseed=11)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=8,bin_width=1.5)
        m.centiles_normative_model()
        assert np.sum(m.data.Centiles) == 446

    def test_get_masks(self):
        a = np.array(list(range(6)))
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        ctr, prob = m._get_masks()
        assert a[ctr].shape[0] == 5
        assert a[prob][0] == 3

    def test_get_masks_all_CON(self):
        a = np.array(list(range(12)))
        data = generate_data(randseed=1)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        ctr, prob = m._get_masks()
        assert a[ctr].shape[0] == 12
        assert a[prob].shape[0] == 0

    def test_get_conf_mat(self):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        conf_mat = m._get_conf_mat()
        assert conf_mat.shape[0] == 6
        assert conf_mat.shape[1] == 3
        for i in range(3):
            assert not isinstance(conf_mat[0, i], str)

    def test_use_approx_auto_small(self):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        assert m._use_approx(method='auto') == False

    def test_use_approx_auto_big(self):
        data = generate_data(randseed=3,sample_size=1000)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        assert m._use_approx(method='auto') == True

    def test_use_approx_approx(self):
        data = generate_data(randseed=3,sample_size=1000)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        assert m._use_approx(method='approx') == True
    
    def test_use_approx_exact(self):
        data = generate_data(randseed=3,sample_size=2000)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        with pytest.warns(Warning) as record:
            use_approx = m._use_approx(method='exact')
        assert len(record) == 1
        assert record[0].message.args[0] == "Exact GP model with over 2000 data points requires large amounts of time and memory, continuing with exact model."
        assert use_approx == False

    def test_gp_normative_model(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gp_normative_model()
        assert 'GP_pred' in m.data.columns
        assert math.isclose(0,m.data['GP_residuals'].mean(),abs_tol=0.5)

    def test_homo_res(self):
        data = dataset_homo()
        m = pynm.PyNM(data,'score','train_sample',['x'])
        with pytest.warns(None) as record:
            m.gp_normative_model(method='exact')
        assert len(record) == 0

    def test_nongaussian_res(self):
        data = dataset_skew()
        m = pynm.PyNM(data,'score','train_sample',['x'])
        with pytest.warns(Warning) as record:
            m.gp_normative_model(method='exact')
        assert len(record) == 1
        assert record[0].message.args[0] == "The residuals are not Gaussian!"

    def test_het_res(self):
        data = dataset_het()
        m = pynm.PyNM(data,'score','train_sample',['x'])
        with pytest.warns(Warning) as record:
            m.gp_normative_model(method='exact')
        assert len(record) == 1
        assert record[0].message.args[0] == "The residuals are heteroskedastic!"
    
    def test_rmse(self):
        y_true = np.array([1,2,3,4,5])
        y_pred = np.array([0,1,3,5,2])
        assert RMSE(y_true,y_pred) == np.sqrt(2.4)

    def test_smse(self):
        y_true = np.array([1,2,3,4,5])
        y_pred = np.array([0,1,3,5,2])
        assert SMSE(y_true,y_pred) == np.sqrt(2.4)/np.sqrt(2)
    
    def test_msll(self):
        y_true = np.array([1,2,3,4,5])
        y_pred = np.array([0,1,3,5,2])
        sigma = np.array([1,2,1,2,1])
        msll = 0.5*np.log(2*np.pi*np.array([1,4,1,4,1])) + np.array([1/2,1/8,0,1/8,9/2]) - np.array([4,1,0,1,4])/(2*np.sqrt(2))
        assert MSLL(y_true,y_pred,sigma) == np.mean(msll)
    
    def test_report(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        m.loess_normative_model()
        #m.gp_normative_model()
        #m.gamlss_normative_model()
        m.report()

@patch("matplotlib.pyplot.show")
class TestPlot:
    def test_plot_default(self,mock_patch):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        m.plot()
        assert True
    
    def test_plot_default_two_models(self,mock_patch):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        m.loess_normative_model()
        assert m.plot() is None
    
    def test_plot_default_no_models(self,mock_patch):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        with pytest.warns(Warning) as record:
            m.plot()
    
    def test_plot_valid_subset(self,mock_patch):
        subset = ['Centiles','LOESS']
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        m.loess_normative_model()
        assert m.plot(kind=subset) is None
    
    def test_plot_invalid_subset1(self,mock_patch):
        subset = ['Centiles',None]
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        with pytest.raises(ValueError):
            m.plot(kind=subset)
    
    def test_plot_invalid_subset2(self,mock_patch):
        subset = ['Centiles','GAMLSS']
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        with pytest.raises(KeyError):
            m.plot(kind=subset)
    
    def test_plot_res_default(self,mock_patch):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        assert m.plot_res() is None
    
    def test_plot_res_default_two_models(self,mock_patch):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        m.loess_normative_model()
        assert m.plot_res() is None
    
    def test_plot_res_default_no_models(self,mock_patch):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        with pytest.raises(ValueError):
            m.plot_res()
    
    def test_plot_res_valid_subset(self,mock_patch):
        subset = ['Centiles','LOESS']
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        m.loess_normative_model()
        assert m.plot_res(kind=subset) is None
    
    def test_plot_res_invalid_subset1(self,mock_patch):
        subset = ['Centiles',None]
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        with pytest.raises(ValueError):
            m.plot_res(kind=subset)
    
    def test_plot_res_invalid_subset2(self,mock_patch):
        subset = ['Centiles','GAMLSS']
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        with pytest.raises(ValueError):
            m.plot_res(kind=subset)
    
    def test_plot_z_default(self,mock_patch):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        assert m.plot_z() is None
    
    def test_plot_z_default_two_models(self,mock_patch):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        m.loess_normative_model()
        assert m.plot_z() is None
    
    def test_plot_z_default_no_models(self,mock_patch):
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        with pytest.raises(ValueError):
            m.plot_z()
    
    def test_plot_z_valid_subset(self,mock_patch):
        subset = ['Centiles','LOESS']
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        m.loess_normative_model()
        assert m.plot_z(kind=subset) is None
    
    def test_plot_z_invalid_subset1(self,mock_patch):
        subset = ['Centiles',None]
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        with pytest.raises(ValueError):
            m.plot_z(kind=subset)
    
    def test_plot_z_invalid_subset2(self,mock_patch):
        subset = ['Centiles','GAMLSS']
        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.centiles_normative_model()
        with pytest.raises(ValueError):
            m.plot_z(kind=subset)

class TestApprox:
    def test_svgp_init(self):
        from pynm.models.approx import SVGP

        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        conf_mat = m._get_conf_mat()
        score = m._get_score()

        X_train,X_test,y_train,y_test = train_test_split(conf_mat, score)
        svgp = SVGP(X_train,X_test,y_train,y_test)
    
    def test_svgp_train(self):
        from pynm.models.approx import SVGP

        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        conf_mat = m._get_conf_mat()
        score = m._get_score()

        X_train,X_test,y_train,y_test = train_test_split(conf_mat, score)
        svgp = SVGP(X_train,X_test,y_train,y_test)
        svgp.train(num_epochs = 2)

        assert len(svgp.loss) == 2
    
    def test_svgp_predict(self):
        from pynm.models.approx import SVGP

        data = generate_data(randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        conf_mat = m._get_conf_mat()
        score = m._get_score()

        X_train,X_test,y_train,y_test = train_test_split(conf_mat, score)
        svgp = SVGP(X_train,X_test,y_train,y_test)
        svgp.train(num_epochs = 2)
        means,sigmas = svgp.predict()

        assert means.size(0) == y_test.shape[0]
        assert sigmas.size(0) == y_test.shape[0]

    def test_svgp_model(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gp_normative_model(method='approx')

        assert 'GP_pred' in m.data.columns
        assert math.isclose(0, m.data['GP_residuals'].mean(), abs_tol=0.5)

class TestGAMLSS:
    def test_get_r_formulas(self):
        from pynm.models import gamlss

        g = gamlss.GAMLSS(mu='score ~ 1')
        mu,sigma,_,_ = g._get_r_formulas('score ~ cs(age) + site',None,None,None)
        #assert not isinstance(mu,str)
        assert mu == 'score ~ cs(age) + site'
        assert sigma == '~ 1'

    def test_gamlss(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gamlss_normative_model(mu='score ~ cs(age)',sigma='~ age + site',tau='~ c(sex)')
        assert 'GAMLSS_pred' in m.data.columns
    
    def test_gamlss_smse(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gamlss_normative_model(mu='score ~ cs(age)',sigma='~ age + site',tau='~ c(sex)')
        assert m.SMSE_GAMLSS > 0
    
    def test_gamlss_default_formulas(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gamlss_normative_model()
        assert 'GAMLSS_pred' in m.data.columns
    
    def test_gamlss_invalid_init(self):
        from pynm.models import gamlss
        
        with pytest.raises(ValueError):
            gamlss.GAMLSS()
    
    def test_gamlss_nan_issue(self):
        df = generate_data(n_sites=4,sample_size=35,randseed=650)
        #Initialize pynm w/ data and confounds
        m = pynm.PyNM(df,'score','group',['age','c(sex)','c(site)'])
        m.loess_normative_model()
        m.centiles_normative_model()
        m.gamlss_normative_model(mu='score ~ ps(age) + c(sex) + c(site)',sigma = '~ age',family='SHASHo2')

    def test_gamlss_random_effect(self):
        df = generate_data(n_sites=4,sample_size=35,randseed=650)
        #Initialize pynm w/ data and confounds
        m = pynm.PyNM(df,'score','group',
                confounds = ['age','c(sex)','c(site)'])
        m.gamlss_normative_model(mu='score ~ ps(age) + c(sex) + random(as.factor(site))',sigma = '~ ps(age)',family='SHASHo2',method='mixed(10,50)')
    
    def test_gamlss_random_effect_not_converged(self):
        #TODO: Force example where algorithm not converged warning gets thrown
        df = generate_data(n_sites=4,sample_size=35,randseed=650)
        #Initialize pynm w/ data and confounds
        m = pynm.PyNM(df,'score','group',
                confounds = ['age','c(sex)','c(site)'])
        m.gamlss_normative_model(mu='score ~ ps(age) + c(sex) + random(as.factor(site))',sigma = '~ ps(age)',family='SHASHo2',method='RS')
    
    def test_gamlss_bad_formula(self):
        df = generate_data(n_sites=4,sample_size=35,randseed=650)
        #Initialize pynm w/ data and confounds
        m = pynm.PyNM(df,'score','group',
                confounds = ['age','c(sex)','c(site)'])
        with pytest.raises(ValueError):
            m.gamlss_normative_model(mu='score ~ xxx(age) + c(sex) + c(site)',family='SHASHo2')

class TestCV:
    def test_cv_1_loess(self):
        data = generate_data(randseed=11)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=8,bin_width=1.5)
        m.loess_normative_model()
        assert math.isclose(2.3482, np.sum(m.data.LOESS_z), abs_tol=0.00001)
    
    def test_cv_3_loess(self):
        data = generate_data(n_sites=1,sample_size=100,randseed=650)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=8,bin_width=1.5)
        m.loess_normative_model(cv_folds=3)
        assert not np.isnan(m.RMSE_LOESS)
        assert not np.isnan(m.SMSE_LOESS)
    
    def test_cv_1_centiles(self):
        data = generate_data(randseed=11)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=8,bin_width=1.5)
        m.centiles_normative_model()
        assert np.sum(m.data.Centiles) == 446
    
    def test_cv_3_centiles(self):
        data = generate_data(n_sites=1,sample_size=100,randseed=650)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'],bin_spacing=8,bin_width=1.5)
        m.centiles_normative_model(cv_folds=3)
        assert not np.isnan(m.RMSE_Centiles)
        assert not np.isnan(m.SMSE_Centiles)
    
    def test_cv_1_gp(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gp_normative_model()
        assert 'GP_pred' in m.data.columns
        assert math.isclose(0,m.data['GP_residuals'].mean(),abs_tol=0.5)
    
    def test_cv_3_gp(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gp_normative_model(cv_folds=3)
        assert 'GP_pred' in m.data.columns
        assert math.isclose(0,m.data['GP_residuals'].mean(),abs_tol=0.5)
    
    def test_cv_1_svgp(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gp_normative_model(method='approx')

        assert 'GP_pred' in m.data.columns
        assert math.isclose(0, m.data['GP_residuals'].mean(), abs_tol=0.5)
    
    def test_cv_3_svgp(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gp_normative_model(method='approx',cv_folds=3,num_epochs=3)

        assert 'GP_pred' in m.data.columns
        assert math.isclose(0, m.data['GP_residuals'].mean(), abs_tol=0.5)
    
    def test_cv_1_gamlss(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gamlss_normative_model(mu='score ~ cs(age)',sigma='~ age + site',tau='~ c(sex)')
        assert 'GAMLSS_pred' in m.data.columns
    
    def test_cv_3_gamlss(self):
        data = generate_data(sample_size=4, n_sites=2, randseed=3)
        m = pynm.PyNM(data,'score','group',['age','c(sex)','c(site)'])
        m.gamlss_normative_model(mu='score ~ cs(age)',sigma='~ age + site',tau='~ c(sex)',cv_folds=3)
        assert 'GAMLSS_pred' in m.data.columns


import numpy as np

def read_confounds(confounds):
    """ Process input list of confounds.

    Parameters
    ----------
    confounds : list of str
        List of confounds with categorical variables indicated by c(var) ('c' must be lower case).

    Returns
    -------
    list
        List of all confounds without wrapper on categorical variables: c(var) -> var.
    list
        List of only categorical confounds without wrapper.
    """
    categorical = []
    clean_confounds = []
    for conf in confounds:
        if ((conf[0:2] == 'c(') & (conf[-1] == ')')):
            categorical.append(conf[2:-1])
            clean_confounds.append(conf[2:-1])
        else:
            clean_confounds.append(conf)
    return clean_confounds, categorical

def RMSE(y_true,y_pred):
    """Calculates Root Mean Square Error (RMSE).

    Parameters
    ----------
    y_true: array
        True values for response variable.
    y_pred: array
        Predicted values for response variable
    
    Returns
    -------
    float
        RMSE value for inputs.
    """
    return (np.mean((y_true - y_pred)**2))**0.5

def SMSE(y_true,y_pred):
    """Calculates Standardized Mean Square Error (SMSE).

    Parameters
    ----------
    y_true: array
        True values for response variable.
    y_pred: array
        Predicted values for response variable
    
    Returns
    -------
    float
        SMSE value for inputs.
    """
    return (np.mean((y_true - y_pred)**2))**0.5/np.std(y_true)

def MSLL(y_true,y_pred,sigma):
    """Calculates Mean Standardized Log Loss (MSLL).

    Parameters
    ----------
    y_true: array
        True values for response variable.
    y_pred: array
        Predicted values for response variable
    sigma: array
        Standard deviation of predictive distribution.
    
    Returns
    -------
    float
        MSLL value for inputs.
    """
    SLL = (0.5 * np.log(2 * np.pi * sigma**2) +
                    (y_true - y_pred)**2 / (2 * sigma**2) -
                    (y_true - np.mean(y_true))**2 /
                    (2 * np.std(y_true)))
    return np.mean(SLL)